# -*-python-*-

__package__    = "pocketMMap/__init__.py"
__version__    = "0.3"
__author__     = "Aaron Straup Cope"
__url__        = "http://www.aaronland.info/python/pocketMMap"
__date__       = "$Date: 2008/11/25 06:54:36 $"
__copyright__  = "Copyright (c) 2007-2008 Aaron Straup Cope. BSD license : http://www.modestmaps.com/license.txt"

# Hey look !! This is version 0.3 !!!
#
# This means it works but it also means
# there are lots of niggly details and 
# features to work out still. As of this
# writing they include:
#
# proper documentation (see __main__ instead)
#
# clean up some of the stupider variable names
#
# font sizes and the pagination widgets
# (automatic resizing and sanity checking)
#
# consider backporting the pagination widgets
# to not require pycairo
#
# ws-pocketmmap

import Image
import cairo

import ModestMaps
import math

from pocketmod import pocketmod
from pocketMMap.pocketMMarkers import pocketMMarkers

class pocketMMap (pocketmod) :

    def __init__ (self, paper_width, paper_height, paper_margin=0, paper_dpi=144) :

        pocketmod.__init__(self, paper_width, paper_height, paper_margin, paper_dpi=144)
        
        #
                
    	self.bbox = []
        self.zoom = 0

        # to keep track of marker/toc stuff

        self.markers = []        
        self.marker_map = {}
        self.marker_data = {}
        
        # the reverse map maps row-col keys to
        # page numbers

        self.reverse_map = {}

        self.count_rows = 0
        self.count_cols = 0

        # Modest Maps stuff

        self.map_image = None
        self.map_obj = None
        self.map_provider = None

        # leftover and hangers-on (maybe?)
        # for the joiners I think...

        self.cols = 0
        self.rows = 0        
        self.pieces = []
        self.pieces_map = {}

    # ##########################################################

    def draw_pg_number (self, im, page_number) :

        im = pocketmod.draw_pg_number(self, im, page_number)
        
        if self.pieces_map.has_key(page_number) :
            # this is kind of stupid, huh...
        
            pm = self.pil2cairo(im)
            self.draw_joiners(pm, page_number)

            im = self.cairo2pil(pm)

        return im
    
    # ##########################################################
    
    def add_markers (self, lotsofdata) :
        for data in lotsofdata :
            self.add_marker(data)

    # ##########################################################
    
    def add_marker (self, data) :

        required = ('latitude', 'longitude')
        
        for item in required :

            if not data.has_key(item) :
                # error
                return False

        self.markers.append(data)

    # ##########################################################
    
    def load_provider (self, label) :        
        self.map_provider = self.__load_provider(label)

    # please to be part of MM core...
    
    def __load_provider (self, name) :

        if name == 'MICROSOFT_ROAD':
            return ModestMaps.Microsoft.RoadProvider()
        
        elif name == 'MICROSOFT_AERIAL':
            return ModestMaps.Microsoft.AerialProvider()
        
        elif name == 'MICROSOFT_HYBRID':
            return ModestMaps.Microsoft.HybridProvider()
        
        elif name == 'GOOGLE_ROAD':
            return ModestMaps.Google.RoadProvider()
        
        elif name == 'GOOGLE_AERIAL':
            return ModestMaps.Google.AerialProvider()
        
        elif name == 'GOOGLE_HYBRID':
            return ModestMaps.Google.HybridProvider()
        
        elif name == 'YAHOO_ROAD':
            return ModestMaps.Yahoo.RoadProvider()
        
        elif name == 'YAHOO_AERIAL':
            return ModestMaps.Yahoo.AerialProvider()
        
        elif name == 'YAHOO_HYBRID':
            return ModestMaps.Yahoo.HybridProvider()
        
        elif name == 'BLUE_MARBLE':
            return ModestMaps.BlueMarble.Provider()
        
        elif name == 'OPEN_STREET_MAP':
            return ModestMaps.OpenStreetMap.Provider()
        
        else :
            return None
        
    # ##########################################################
    
    def draw (self, bbox, zoom, out=None) :

        self.setup(bbox, zoom)
        self.fetch_map()

        self.plot_markers()
        self.slice_map()

        self.generate_index()

        self.layout()

        if out :
            self.save(out)

    # ##########################################################        
    
    def fetch_map (self) :

        coord, offset = ModestMaps.calculateMapExtent(self.map_provider,
                                                      self.map_width, self.map_height,
                                                      ModestMaps.Geo.Location(self.bbox[0], self.bbox[1]),
                                                      ModestMaps.Geo.Location(self.bbox[2], self.bbox[3]))
        
        dim = ModestMaps.Core.Point(self.map_width, self.map_height)
        self.map_obj = ModestMaps.Map(self.map_provider, dim, coord, offset)            
        
        self.map_image = self.map_obj.draw()
        self.map_image = self.map_image.convert("RGBA")

    # ##########################################################                

    def plot_markers (self) :

        if len(self.markers) == 0 :
            return
        
        i = 1
        
        map_surface = self.pil2cairo(self.map_image)

        bg_c = (.5, 0, 0)
        str_c = (1, 1, 1)
        txt_c = (1, 1, 1)

        #
        
        pm = pocketMMarkers(self.map_obj)
        sorted = pm.reposition_markers(self.markers)

        #
        
        for data in sorted :

            row = int(math.ceil(float(data['y']) / float(self.sect_height)))
            col = int(math.ceil(float(data['x']) / float(self.sect_width)))
            
            self.marker_map[i] = self.mk_key_for_row_col(row, col)
            self.marker_data[i] = data

            args = {
                'x' : data['x'], 'y' : data['y'], 'radius' : 10, 'string' : i,
                'background_c':bg_c, 'stroke_c':str_c, 'text_c' : txt_c,
                'font' : 'Helvetica-bold',
                'anchor_pos':data['anchor_pos'], 'anchor_height':data['anchor_height'],
                }
            
            self.draw_circle_with_string(map_surface, args)

            i += 1
            
        self.map_image = self.cairo2pil(map_surface)

    # ##########################################################
    
    def slice_map (self) :

        #
        # Split the map into multiple pieces
        #
        
        offset_x = 0
        offset_y = 0
        row = 0
        col = 0

        # not (1) because we're talking page numbers
        # here and the first chunk of map always
        # appears on page two
        
        i = 2

        self.pages.append(None)
        
        while offset_y < self.map_height :

            while offset_x < self.map_width :

                x1 = offset_x
                y1 = offset_y
                x2 = x1 + self.sect_width
                y2 = y1 + self.sect_height
                
                box = (x1, y1, x2, y2)
                
                sect = self.crop_map_image(box)
                self.pages.append(sect)

                # do not use a zero based count because
                # it makes life hard when we plot markers
                # and have to figure out which page/section
                # they live on based on their (x,y) coords

                key_row = row + 1
                key_col = col + 1
                
                reverse_key = self.mk_key_for_row_col(key_row, key_col)
                
                self.pieces_map[i] = (key_row, key_col)            
                self.reverse_map[reverse_key] = i

                offset_x += self.sect_width
                
                col += 1
                i += 1

            self.count_cols = col
            col = 0
            
            row += 1
    
            offset_x = 0
            offset_y += self.sect_height

        self.count_rows = row
        
    # ##########################################################

    def generate_index (self) :

        if len(self.marker_map.keys()) == 0 :
            return

        #
        
        try :
            from pocketindex import pocketindex
        except Exception, e:
            print e
            return
        
        #
        
        toc_data = []

        for marker in self.marker_map.keys() :
            key = self.marker_map[marker]
            page = self.reverse_map[key]
            data = self.marker_data[marker]

            data['value'] = "%s / %s" % (page, marker)
            toc_data.append(data)
        
        idx = pocketindex(self.sect_width, self.sect_height)
        pages = idx.create(toc_data)

        # flip the list so it gets rendered properly
        pages.reverse()

        self.toc_pages = pages

    # ##########################################################
    
    def setup (self, bbox, zoom) :

	self.bbox = bbox
    	self.zoom = zoom
        
        # this piece is copied from wscompose/__init__.py    
        
        offset_lat = (self.bbox[2] - self.bbox[0]) / 2
        offset_lon = (self.bbox[3] - self.bbox[1]) / 2

        lat = self.bbox[0] + offset_lat
        lon = self.bbox[1] + offset_lon
        
        loc = ModestMaps.Geo.Location(lat, lon)
        coordinate = self.map_provider.locationCoordinate(loc)            
        coordinate = coordinate.zoomTo(self.zoom)
        
        coord, offset = ModestMaps.calculateMapCenter(self.map_provider, coordinate)
        dim = ModestMaps.Core.Point(0, 0)
        
        self.map_obj = ModestMaps.Map(self.map_provider, dim, coord, offset)
        
        # this piece is copied from ModestMaps/__init__.py

        sw = ModestMaps.Geo.Location(self.bbox[0], self.bbox[1])
        ne = ModestMaps.Geo.Location(self.bbox[2], self.bbox[3])
        nw = ModestMaps.Geo.Location(ne.lat, sw.lon)
        se = ModestMaps.Geo.Location(sw.lat, ne.lon)
        
        TL = self.map_provider.locationCoordinate(nw).zoomTo(zoom)
        
        # don't need this (note to self when trying to merge in to core...)
        # tiles = ModestMaps.TileQueue()
        
        cur_lon = sw.lon
        cur_lat = ne.lat        
        max_lon = ne.lon
        max_lat = sw.lat
        
        x_off = 0
        y_off = 0
        tile_x = 0
        tile_y = 0
        
        tileCoord = TL.copy()
        
        while cur_lon < max_lon :

            y_off = 0
            tile_y = 0
            
            while cur_lat > max_lat :

                # don't need this (note to self when trying to merge in to core...)
                # tiles.append(ModestMaps.TileRequest(self.map_provider, tileCoord, ModestMaps.Core.Point(x_off, y_off)))
                
                y_off += self.map_provider.tileHeight()
                
                tileCoord = tileCoord.down()
                loc = self.map_provider.coordinateLocation(tileCoord)
                cur_lat = loc.lat

                tile_y += 1
                
            x_off += self.map_provider.tileWidth()            
            cur_lat = ne.lat
            
            tile_x += 1
            tileCoord = TL.copy().right(tile_x)

            loc = self.map_provider.coordinateLocation(tileCoord)
            cur_lon = loc.lon

        self.map_width = int(self.map_provider.tileWidth() * tile_x)
        self.map_height = int(self.map_provider.tileHeight() * tile_y)

        #
        # Work out the various dimensions
        # Adjust the map_width and map_height variables
        # as needed to ensure coverage for all
        # the sections and that rows have an
        # even number of columns
        #
        
        # flip!
        #
        # I can't remember why I thought this was a good idea
        # but it is the way it works now - something to do with
        # always generating the source map image in landscape
        # mode...I think

        self.width = int(self.paper_height * self.paper_dpi)
        self.height = int(self.paper_width * self.paper_dpi)
        self.margin = int(self.paper_margin * self.paper_dpi)

        self.sect_width = int((self.width - (self.margin * 4)) / 4)
        self.sect_height = int((self.height - (self.margin * 2)) / 2)

        mod_w = (self.map_width % self.sect_width)
        mod_h = (self.map_height % self.sect_height)

        #
        # Adjust adjust adjust
        #

        self.map_width += mod_w
        self.map_height += mod_h

        if ((self.map_width / self.sect_width) % 2) :
            self.map_width += self.sect_width

        count_w = (self.map_width / self.sect_width)
        count_h = (self.map_height / self.sect_height)

        self.cols = count_w
        self.rows = count_h
        
        self.map_width = count_w * self.sect_width
        self.map_height = count_h * self.sect_height

        self.sect_map = {
            'a' : {'rotate' : Image.ROTATE_270, 'x' : 0, 'y' : 0},
            'b' : {'rotate' : Image.ROTATE_90, 'x' : self.sect_height, 'y' : 0},
            
            'c' : {'rotate' : Image.ROTATE_270, 'x' : 0, 'y' : self.sect_width},
            'd' : {'rotate' : Image.ROTATE_90, 'x' : self.sect_height, 'y' : self.sect_width}, 
            
            'e' : {'rotate' : Image.ROTATE_270, 'x' : 0, 'y' : 2 * self.sect_width},
            'f' : {'rotate' : Image.ROTATE_90, 'x' : self.sect_height, 'y' : 2 * self.sect_width},
            
            'g' : {'rotate' : Image.ROTATE_270, 'x' : 0, 'y' : 3 * self.sect_width},
            'h' : {'rotate' : Image.ROTATE_90, 'x' : self.sect_height, 'y' : 3 * self.sect_width}
            }

        # profit!

    # ##########################################################

    def draw_joiners (self, surface, pg_number) :

        (row, col) = self.pieces_map[ pg_number ]
                
        n = None
        s = None
        e = None
        w = None
        
        # north
    
        if row != 1 :
            n = ((row - 1), col)
        
        # south

        if row != self.count_rows :
            s = ((row + 1), col)
        
        # west

        if col != 1 :
            if pg_number % 2 == 0:            
                w = (row, (col - 1))

        # east

        if col != self.count_cols :
            if pg_number % 2 != 0:
                e = (row, (col + 1))

        #
    
        width = surface.get_width()
        height = surface.get_height()
    
        if n :
            x = int(width / 2)
            y = 30
            self.draw_joiner(surface, x, y, n)

        if s :
            x = int(width / 2)
            y = height - 30
            self.draw_joiner(surface, x, y, s)

        if e :
            x = width - 30
            y = int(height / 2)
            self.draw_joiner(surface, x, y, e)

        if w :
            x = 30
            y = int(height / 2)
            self.draw_joiner(surface, x, y, w)

    # ##########################################################        

    def draw_circle_with_string (self, surface, args) :

        bg_c = (255, 255, 255)
        str_c = (0, 0, 0)
        txt_c = (0, 0, 0)
        
        if args.has_key('background_c'):
            bg_c = args['background_c']

        if args.has_key('stroke_c'):
            str_c = args['stroke_c']

        if args.has_key('text_c'):
            txt_c = args['text_c']

        x = args['x']
        y = args['y']
        r = args['radius']
        a = 0

        if args.has_key('anchor_height') :
            a  = args['anchor_height']

        if a != 0 :
            y = y - (a + int(args['radius'] / 2))

        #
        
        ctx = cairo.Context(surface)        
        ctx.move_to(x, y)
        ctx.arc(x, y, r, 0, 360)
        ctx.set_source_rgb(bg_c[0], bg_c[1], bg_c[2])
        ctx.fill()

        if a != 0 :
            ctx = cairo.Context(surface)        
            ctx.move_to(args['x'], args['y'])
            ctx.set_source_rgb(bg_c[0], bg_c[1], bg_c[2])            
            ctx.set_line_width(2)
            ctx.line_to(args['x'], (args['y'] - a))
            ctx.stroke()
        
        ctx = cairo.Context(surface)
        ctx.arc(x, y, r, 0, 360)        
        ctx.set_source_rgb(str_c[0], str_c[1], str_c[2])
        ctx.set_line_width(2)
        ctx.stroke()

        ctx = cairo.Context(surface)                                
        ctx.set_source_rgb(txt_c[0], txt_c[1], txt_c[2])    
        
	offset_x, offset_y = self.offsets_for_page_id(ctx, args['string'])
                
        x = x + offset_x
        y = y - offset_y

        font_face = "Helvetica Bold"
        font_size = 14

        ctx.select_font_face(font_face, cairo.FONT_SLANT_NORMAL, cairo.FONT_WEIGHT_BOLD)
        # ctx.set_font_size(font_size)

        ctx.move_to(x, y)
        ctx.show_text(unicode(args['string']))

    # ##########################################################        
    
    def draw_joiner(self, surface, x, y, pos) :

        key = self.mk_key_for_row_col(pos[0], pos[1])
        pg_number = self.reverse_map[key]

        args = {'x' : x, 'y' : y, 'radius' : 10, 'string' : pg_number}
        
        self.draw_circle_with_string(surface, args)

    # ##########################################################
    
    def crop_map_image(self, box) :
        return self.map_image.crop(box)

    # ##########################################################
    
    def mk_key_for_row_col (self, row, col) :
        return "%s-%s" % (row, col)

    # ##########################################################                
    
if __name__ == '__main__' :

    height = 11
    width = 8.5
    margin = .25
    dpi = 144

    bbox = (45.482882,-73.619899,45.532687,-73.547801)
    # bbox = (40.6977996826,-73.9747772217,40.7119102478,-73.9369964600)
    zoom = 16

    out = "example.pdf"
    
    pm = pocketMMap(height, width, margin, dpi)
    pm.load_provider('MICROSOFT_ROAD')
    pm.draw(bbox, zoom)
    pm.save(out)
